/*
 * JVPanel.java
 *
 * Created on 15. September 2003, 08:11
 */

package org.jvisu.panel;

import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.beans.*;
import java.net.*;
import java.io.*;
import javax.swing.*;
import org.jvisu.base.*;

/**
 *  <p>package:    org.jvisu.panel</p>
 *  <p>file:       JVPanel.java</p>
 * $Id: JVPanel.java,v 1.12 2004/04/19 08:20:55 tgl Exp $
 *
 *  <p>Copyright 2003 Thomas Gleixner <tglx@linutronix.de></p>
 *
 * <p>This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.</p>
 * </p>
 *
 * Change Log:
 * </p>
 * <ul>
 * <li> 1.0 First implementation
 * <li> 1.2 Changes due to new variable and communication handling
 * <li> 1.3 Support absolute component placing with respect to the
 *          fact, that array based component creation must be possible.
 * <li> 1.4 Simplify global settings and support zoomfactor
 * </ul>
 * <p>
 * </p>
 *  The purpose of JVPanel is to provide a XML configurable component, which
 *  can hold one or more JVisu components and standard swing components like
 *  labels. Other components, which need user interaction should not be
 *  used, as at the moment no code can be associated with user actions.
 *
 *  The panel can be used to design complete subsets of an HMI by adding and
 *  layouting the interface with the JVPanelCfg application and store it in
 *  a XML file for runtime configuration.
 *
 * @author Thomas Gleixner <tglx@linutronix.de>
 * @version 1.4
 */
public abstract class JVPanel extends org.jvisu.base.JVBean implements PropertyChangeListener, ContainerListener {
    
    /** The status string */
    protected String                  status;
    /** the background image file address. The address is
     *  relative to base, if base is != null. Else it is
     *  treated as an absolute reference.
     */
    private String                  imageFile;
    /** unscaled background Image */
    protected Image                 bgImage;
    /** scaled background Image */
    private BufferedImage           scaledBgImage;
    /** Set the comm port for all subcomponents */
    protected boolean               setport = true;
    /** Automatic size adjustment ?*/
    protected boolean               autosize = false;
    
    /** Creates a new instance of JVPanel */
    public JVPanel() {
        /* Be informend, when a component is added or removed */
        addContainerListener(this);
    }
    
    /** Property change listener
     *  @param evt the property event
     */
    public void propertyChange(PropertyChangeEvent evt) {
        super.propertyChange(evt);
        String key = evt.getPropertyName();
        
        /* These are local events */
        if (evt.getSource() == this) {
            if (key.equals("base") || key.equals("imageFile")) {
                readImageFile();
            }
            /* Try to inform the child components about changes 
             * to the stuff, which is set globally.
             * ????????
             * The child components could use a propertyChangeListener too,
             * but this works only when the Panel is already visible.
             * I don't know why, but I found no way to make it work
             * before the panel was displayed 
             */
            if (key.equals("base"))
                setChildProp(key, getBase());
            if (key.equals("server"))
                setChildProp(key, getServer());
            if (key.equals("commtype"))
                setChildProp(key, getCommtype());
            if ((key.equals("port") || key.equals("globalPortSetting")) && setport)
                setChildProp("port", Integer.toString(getPort()));
            if (key.equals("editEnabled"))
                setChildProp(key, isEditEnabled() ? "true" : "false");
            if (key.equals("zoomfactor"))
                setChildProp(key, Double.toString(getZoomfactor()));
        } else {
            /** Child events */
            if (key.equals("preferredSize") || key.equals("designpos"))
                this.revalidate();
        }
    }

    /** Inform all subcomponents about global changes
     */
    private void setChildProp(String name, String value) {
        for (int i = 0; i < this.getComponentCount(); i++) {
            try {
                JComponent c = (JComponent) this.getComponent(i);
                JVBasePanel.setProperty(c, name, value);
            } catch (Exception e) { }
        }
    }
    
    /** A component was added
     */
    public void componentAdded(ContainerEvent e) {
        if (e.getChild() instanceof JVBaseMembers) {
            e.getChild().addPropertyChangeListener(this);
            ((JVBaseMembers) e.getChild()).setZoomfactor(getZoomfactor());
        }
        try {
            JVBean bean = (JVBean) e.getChild();
            bean.setBase(getBase());
            bean.setServer(getServer());
            if (setport)
                bean.setPort(getPort());
            bean.setCommtype(getCommtype());
            bean.setEditEnabled(isEditEnabled());
        } catch (Exception ex) { 
        }
    }

    /** A component was removed
     */
    public void componentRemoved(ContainerEvent e) {
        e.getChild().removePropertyChangeListener(this);
    }

    /** Revalidate the layout of the component.
     */
    public void revalidate() {
        recalcLayout();
        super.revalidate();
    }
    
    /** Optimize the panel size and set components positions,
     *  if the layout is SpringLayout
     */
    protected void recalcLayout() {
        SpringLayout layout;
        try {
            layout = (SpringLayout) getLayout();
            if (layout == null)
                return;
        } catch (ClassCastException e) {
            /* TODO: clear preferred size  ??? */
            return;
        }
        
        /* Take border into account */
        int insx = 0, insy = 0;
        javax.swing.border.Border border = getBorder();
        if (border != null) {
            Insets ins = getInsets();
            insx += ins.left;
            insy += ins.top;
        }

        int x = 0, y = 0;
        for (int i = 0; i < getComponentCount(); i++) {
            Component c = getComponent(i);
            
            try {
                JVBaseMembers comp =(JVBaseMembers) c;
                Point p = comp.getDesignpos();
                double px = ((double) p.x) * zoomfactor;
                double py = ((double) p.y) * zoomfactor;
                layout.putConstraint(SpringLayout.WEST,(Component) comp, insx + (int)px, SpringLayout.WEST, this);
                layout.putConstraint(SpringLayout.NORTH,(Component) comp, insy + (int)py, SpringLayout.NORTH, this);
            } catch (ClassCastException e) {}

            Dimension size = c.getPreferredSize();
            SpringLayout.Constraints constr = layout.getConstraints(c);
            if ((size.width + constr.getX().getValue()) > x)
                x = size.width + constr.getX().getValue();
            if ((size.height + constr.getY().getValue()) > y)
                y = size.height + constr.getY().getValue();
        }

        /* Take border into account */
        if (border != null) {
            Insets ins = getInsets();
            x += ins.left + ins.right;
            y += ins.top + ins.bottom;
        }
        
        if (autosize && bgImage == null) {
            Dimension actsize = getPreferredSize();
            setPreferredSize(new Dimension(x, y));
        }
    }    
    
    /** Paint the component, if a background image is given
     *  @param g    the Graphics context to draw
     */
    public void paintComponent(Graphics g) {
        if (status != null) {
            g.setColor (Color.RED);
            g.fillRect (0, 0, getSize().width, getSize().height);
            g.setColor (Color.BLACK);
            g.drawString (status, 0, 20);
            return;
        }
        if (bgImage == null)
            super.paintComponent(g);
        else {
            if (scaledBgImage == null ||
            scaledBgImage.getHeight() != this.getHeight() ||
            scaledBgImage.getWidth() != this.getWidth()) {
                
                scaledBgImage = new BufferedImage(getWidth(), getHeight(),java.awt.image.BufferedImage.TYPE_INT_ARGB);
                scaledBgImage.getGraphics().drawImage(
                bgImage.getScaledInstance(this.getWidth(), this.getHeight(), java.awt.Image.SCALE_FAST), 0, 0, null);
            }
            g.drawImage(scaledBgImage, 0, 0, null);
        }
    }
    
    /** Return an subimage of the background image, which represents the
     *  components rectangle
     */
    public BufferedImage getComponentImage(JComponent comp) {
        BufferedImage img = scaledBgImage;
        if (img == null) {
            if (getParent() instanceof JVPanel)
                img = ((JVPanel)getParent()).getComponentImage(this);
        }
        try {
            return img.getSubimage(comp.getX(), comp.getY(), comp.getWidth(), comp.getHeight());
        } catch (Exception e) {
            return null;
        }
    }
    
    /** Try to read the image file */
    private void readImageFile() {
        scaledBgImage = null;
        try {
            URL url;
            if (base == null)
                url = new URL(imageFile);
            else
                url = new URL(new URL(base), imageFile);
            //System.out.println (url);
            bgImage= javax.imageio.ImageIO.read(url);
            if (autosize)
                this.setDesignsize(new Dimension( bgImage.getWidth(null), bgImage.getHeight(null)));
        } catch (Exception e) {
            bgImage = null;
        }
    }

    /** Get the autosize property
     *  @return true, if autosize is enabled
     */
    public boolean isAutoSized() {
        return autosize;
    }

    /** Set the autosize property
     *  @param auto true, if autosize is enabled
     */
    public void setAutoSized(boolean auto) {
        autosize = auto;
    }
    
    /** Get the address of the background image file
     * @return background image file address
     */
    public String getImageFile() {
        return this.imageFile;
    }
    
    /** Set the background image file address
     * @param imgfile new background image file address
     */
    public void setImageFile(String imgfile) {
        String old = imageFile;
        imageFile = imgfile;
        firePropertyChange("imageFile", old, imageFile);
    }
    
    /** Get the ability of this component to have subcomponents
     * @return true, if subcomponents are allowed
     */
    public boolean isContainer() {
        return true;
    }

    /** Get the ability of this component to set the
     *  communication port for all its subcomponents
     * @return true, if port setting is active
     */
    public boolean isGlobalPortSetting() {
        return setport;
    }

    /** Set the ability of this component to set the
     *  communication port for all its subcomponents
     * @param set true, if port setting is active
     */
    public void setGlobalPortSetting(boolean set) {
        boolean old = setport;
        setport = set;
        firePropertyChange("globalPortSetting", old, setport);
    }
}
