package org.jvisu.panel.configurator;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.beans.*;
import javax.swing.*;
import javax.swing.border.*;
import org.jvisu.util.*;
import org.jvisu.panel.configurator.editors.*;
import org.jvisu.panel.configurator.events.*;

/** Panel property editor
 *  <p>package:    org.jvisu.panel</p>
 *  <p>file:       JVPanelPropertyEditor.java</p>
 * $Id: JVPanelPropertyEditor.java,v 1.4 2004/04/19 08:20:55 tgl Exp $
 *
 *  <p>Copyright 2003 Thomas Gleixner <tglx@linutronix.de></p>
 *
 * <p>This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.</p>
 * </p>
 *
 * Change Log:
 * </p>
 * <ul>
 * <li> 1.0 First implementation
 * </ul>
 * <p>
 * </p>
 * JVPanelPanelPropertyEditor is a component to edit the properties of an
 * JVPanel visualization element.
 *  When a component is assigned the component properties are read via
 *  introspection. The non default properties can be read out in form of
 *  TreeItems, which are used later to produce the visualization description
 *  in XML
 *
 * @author Thomas Gleixner <tglx@linutronix.de>
 * @version 1.3
 */
public class JVPanelPropertyEditor extends javax.swing.JScrollPane {
    
    /** Tabbed pane for property categories */
    private JTabbedPane         pTabs;
    /** Component, which has to be introspected */
    private JComponent          editComponent;
    /** Component properties */
    private ArrayList           props;
    /** Standard panel size */
    private Dimension           size = new Dimension (80, 20);
    /** Standard panel font */
    private Font                font = new Font("Dialog", Font.PLAIN, 11);
    /** Property enable selector */
    private boolean             showAllProperties;
    /** color chooser */
    private JColorChooser       colorChooser = new JColorChooser();
    /** Color chooser dialog */
    private JDialog             colorDlg;
    /** Font chooser dialog */
    private JVFontChooserDlg    fontDlg;
    /** Dialog closed with OK */
    private boolean             dialogOK;
    /** Editor event support */
    private JVEditEventSupport  eventSupport;
    
    /** Creates a new instance of JVPanelPropertyEditor */
    public JVPanelPropertyEditor() {
        pTabs = new JTabbedPane();
        pTabs.setFont(font);
        this.setViewportView(pTabs);
        eventSupport = new JVEditEventSupport();
        
        PropertyEditorManager.registerEditor(Dimension.class, JVDimensionEditor.class);
        PropertyEditorManager.registerEditor(Point.class, JVPointEditor.class);

        /* Provide dialogs for color and font */
        colorDlg = JColorChooser.createDialog(this, "Select color", true, colorChooser, 
            new ActionListener() {
                public void actionPerformed (ActionEvent e) {
                    dialogOK = true;
                }
            } , null);
        
        fontDlg = new JVFontChooserDlg(null, font,
            new ActionListener() {
                public void actionPerformed (ActionEvent e) {
                    dialogOK = true;
                }
            });
    }

    /** Add an listener 
     *  @param listener Listener to add to the listener list
     */
    public void addEditEventListener(JVEditEventListener listener) {
        eventSupport.addEditEventListener(listener);
    }

    /** Remove an listener 
     *  @param listener Listener to remove from the listener list
     */
    public void removeEditEventListener(JVEditEventListener listener) {
        eventSupport.removeEditEventListener(listener);
    }
    
    /** A method for the property cellpanels to fire a property
     *  change event.
     */
    /** Fire an editor event 
     *  @param evt the new editor event, which should be fired
     */
    public void fireEditEvent (JVEditEvent evt) {
        eventSupport.fireEditEvent(evt);
    }    
    
    /** Return if a custom dialog for this class is available
     *  @param clas class to get a custom editor
     *  @return true, if a dialog is available
     */
    public boolean  hasCustomDialog (Class clas) {
        return (clas.isAssignableFrom(Color.class) ||
            clas.isAssignableFrom(Font.class));
    }
    
    /** Show an custom dialog for a given class
     *  The dialog is initialized with the given object
     *  If the dialog closes with ok, the modified object is returned
     *  else null
     *  @param clas class to get a custom editor
     *  @param obj  object to intialize custom editor
     *  @return modified object or null, if no dialog is found
     */
    public synchronized Object showCustomDialog (Class clas, Object obj) {
        dialogOK = false;
        Object modified = null;
        if (clas.isAssignableFrom(Color.class)) {
            colorChooser.setColor((Color) obj);
            colorDlg.show();
            modified = (Object) colorChooser.getColor();
        }
        if (clas.isAssignableFrom(Font.class)) {
            fontDlg.setFont((Font) obj);
            fontDlg.show();
            modified = (Object) fontDlg.getFont();
        }
        return dialogOK ? modified : null;
    }

    /** A class to host the property desriptors
     */
    private class JPropPane extends JSplitPane {
        
        /* The descripton panel */
        protected JPanel pDescr;
        /* The editor panel */
        protected JPanel pEdit;
        
        /* Create a new instance
         **/
        public JPropPane () {
            pDescr = new JPanel();
            pDescr.setLayout(new BoxLayout(pDescr, BoxLayout.Y_AXIS));
            pEdit = new JPanel();
            pEdit.setLayout(new BoxLayout(pEdit, BoxLayout.Y_AXIS));
            Dimension size = new Dimension(100,100);
            pDescr.setMinimumSize(size);
            pEdit.setMinimumSize(size);
            
            setOrientation(JSplitPane.HORIZONTAL_SPLIT);
            this.setLeftComponent(pDescr);
            this.setRightComponent(pEdit);
            setDividerSize(2);
        }
        
        /** Remove all components from the panels
         */
        public void removePanelContent() {
            pDescr.removeAll();
            pDescr.revalidate();
            pDescr.repaint();
            pEdit.removeAll();
            pEdit.revalidate();
            pEdit.repaint();
        }
        
        /** Notification, that the pane is added to a displayable context.
         *  Set the properties, which can be set in displayable context only
         */
        public void addNotify() {
            super.addNotify();
            this.setDividerLocation(0.7);
            setResizeWeight(0.9);
        }
    }
    
    
    /** Find a corresponding tab for this category.
     *  If none is found create a new one
     */
    private JPropPane getTab(String cat) {
        
        for (int i = 0; i < pTabs.getTabCount(); i++) {
            if (pTabs.getTitleAt(i).equalsIgnoreCase(cat)) {
                pTabs.setEnabledAt(i, true);
                return (JPropPane) pTabs.getComponentAt(i);
            }
        }
        // Create a new one and add it to the tabbed pane
        JPropPane p = new JVPanelPropertyEditor.JPropPane();
        pTabs.add(cat, p);
        return p;
    }
    
    
    /** Try to read the component */
    private void readComponent(JComponent comp) {
        
        int i;
        
        /* Clear all stuff in the panels */
        for (i = 0; i < pTabs.getTabCount(); i++) {
            pTabs.setEnabledAt(i, false);
            ((JPropPane) pTabs.getComponentAt(i)).removePanelContent();
        }
        
        if (comp == null) {
            System.out.println ("No component");
            return;
        }
        if (props == null) {
            System.out.println ("No properties");
            return;
        }

        // Build the property panels 
        for (i = 0; i < props.size(); i++) {
            JVPalettePanel.PaletteItemProperty prop;
            prop = (JVPalettePanel.PaletteItemProperty) props.get(i);
            
            if (prop.isFixed())
                continue;                            
            if (prop.isExpert())
                continue;                            

            JPropPane p = getTab (prop.getCategory());
            /* Build the descriptor */
            JVDescrCellPanel desc = new JVDescrCellPanel(prop.toString(), size, font);
            p.pDescr.add(desc);
            JVPropertyCellPanel edit = new JVPropertyCellPanel(comp, prop, size, font, this);
            p.pEdit.add(edit);
        }
        revalidate();
    }
    
    /** Get the edit component
     *  @return comp the component, which is actually inspected
     */
    public JComponent getEditComponent() {
        return editComponent;
    }
    
    /** Set the edit component
     *  @param comp the component, which has to be inspected
     */
    public void setEditComponent(JComponent comp, ArrayList propArray) {
        if (editComponent != comp) {
            editComponent = comp;
            props = propArray;
            readComponent(comp);
        }
    }
    
    /** Getter for property showAllProperties.
     * @return Value of property showAllProperties.
     */
    public boolean isShowAllProperties() {
        return showAllProperties;
    }
    
    /** Setter for property showAllProperties.
     * @param showAll New value of property showAllProperties.
     *
     */
    public void setShowAllProperties(boolean showAll) {
        showAllProperties = showAll;
    }
    
    /** Main method for standalone testing */
    /*
    public static void main(String args[]) {
        JFrame frame = new JFrame();
        final JVPanelPropertyEditor p = new JVPanelPropertyEditor();
        frame.getContentPane().setLayout(new BoxLayout(frame.getContentPane(), BoxLayout.Y_AXIS));
        frame.getContentPane().add(p);

        org.jvisu.base.JVBool pf = new org.jvisu.base.JVBool();
        p.setEditComponent(pf);
        
        JButton b = new JButton("test");
        b.addActionListener( new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                TreeItem it = new TreeItem("component");
                p.updateComponentItem(it);
                ItemToXML xml = new ItemToXML();
                it.printFollow(xml,false);
                try {
                    System.out.println(xml.GetString());
                } catch (Exception e) {System.out.println(e); }
            }
        });
        frame.getContentPane().add(b);
        frame.pack();
        frame.show();
    }
     */
}
