package org.jvisu.panel.configurator;

import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;

/**
 *  <p>package:    org.jvisu.panel</p>
 *  <p>file:       JVPanelTree.java</p>
 * $Id: JVPanelTree.java,v 1.5 2004/04/19 08:20:55 tgl Exp $
 *
 *  <p>Copyright 2003 Thomas Gleixner <tglx@linutronix.de></p>
 *
 * <p>This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.</p>
 * </p>
 *
 * Change Log:
 * </p>
 * <ul>
 * <li> 1.0 First implementation
 * </ul>
 * <p>
 * </p>
 *
 *  JVPanelTree was built with the JVPanelConfigurator in mind. But it is a 
 *  real universal tree storage model. It can be used for various tree based
 *  configuration tasks.
 *
 * @author Thomas Gleixner <tglx@linutronix.de>
 * @version 1.3
 */
public class JVPanelTree extends javax.swing.JScrollPane {

    /** The treeview */
    private JTree                       comptree;    
    /** The topnode */
    private DefaultMutableTreeNode      topnode;
    /** The treemodel */
    private DefaultTreeModel            confmodel; 
    /** The edit popup */
    private JPopupMenu                  editPopup;
    
    /** Creates a new instance of JVPanelTree */
    public JVPanelTree() {
        topnode = new DefaultMutableTreeNode (null);
        confmodel = new DefaultTreeModel(topnode);
        comptree = new JTree(confmodel);
        comptree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
        this.setViewportView(comptree);
        
        comptree.addMouseListener (new java.awt.event.MouseListener() {
            public void mouseClicked (java.awt.event.MouseEvent evt) {
                if (getSelectedNode() != null)
                    if(evt.getButton() == java.awt.event.MouseEvent.BUTTON3)
                        editPopup.show(comptree, evt.getPoint().x,
                            evt.getPoint().y);
            }
            public void mousePressed (java.awt.event.MouseEvent evt) {
            }
            public void mouseReleased (java.awt.event.MouseEvent evt) {
            }
            public void mouseEntered (java.awt.event.MouseEvent evt) {
            }
            public void mouseExited (java.awt.event.MouseEvent evt) {
            }
        });        
    }

    /** Set the popupmenu
     *  @param men  the popup menu
     */
    public void setPopupMenu (JPopupMenu men) {
        editPopup = men;
    }    
    /** Return the tree component 
     *  @return tree
     */
    public JTree getTree() {
        return comptree;
    }
    
    /** Return the topnode of the tree 
     *  @return topnode
     */
    public DefaultMutableTreeNode getTopNode() {
        return topnode;
    }
    
    /** Set the top Object
     *  @param top new top object
     */
    public void setTopObject(Object top) {
        topnode.setUserObject(top);
        confmodel.reload(topnode);
    }

    /** Add a new object to the node, which is identified by the user object
     *  @param parent object, to which should be added
     *  @param comp copmponent, which should be added
     *  @param allowsChildren true, if this component can have children
     */
    public void addObject(Object parent, JVPanelTreeComponent comp, boolean allowsChildren) {
        DefaultMutableTreeNode node = (DefaultMutableTreeNode) findNode(topnode, parent);
        /* Nothing there, nothing to insert */
        if (node == null)
            return;
        /* If this object cannot have children */
        if (!node.getAllowsChildren())
            return;
        DefaultMutableTreeNode child = new DefaultMutableTreeNode (comp);
        confmodel.insertNodeInto(child, node, node.getChildCount());
        confmodel.reload(node);
        this.setSelectedObject(comp.getObject());
    }

    /** Remove a node, identified by the object in user object. 
     *  Release user object reference
     *  @param obj object, which has to be removed
     */
    public void removeObject(Object obj) {
        DefaultMutableTreeNode node = cutObject(obj);
        if (node == null)
            return;
        node.setUserObject(null);
    }

    /** Remove all nodes,
     *  Release user object reference
     */
    public void removeAllObjects() {
        topnode.removeAllChildren();
        confmodel.reload(topnode);
    }
    
    /** Remove a node from the node tree, identified by the userobject, return 
     *  node. Keep userobject reference intact
     *  @param obj object, which has to be removed
     *  @return node, which has been removed
     */
    public DefaultMutableTreeNode cutObject(Object obj) {
        DefaultMutableTreeNode node = (DefaultMutableTreeNode) findNode(topnode, obj);
        if (node == null)
            return null;
        DefaultMutableTreeNode parent = (DefaultMutableTreeNode) node.getParent();
        confmodel.removeNodeFromParent(node);
        confmodel.reload(parent);
        comptree.setSelectionPath(new TreePath(parent.getPath()));
        return node;
    }
    
    /** Get the selected node
     *  @return node, which is selected
     */
    public DefaultMutableTreeNode getSelectedNode() {
        return (DefaultMutableTreeNode) comptree.getLastSelectedPathComponent();
    }

    /** Get the selected Object
     *  @return object, which is selected
     */
    public Object getSelectedObject() {
        DefaultMutableTreeNode node = (DefaultMutableTreeNode) comptree.getLastSelectedPathComponent();
        if (node != null)
            return node.getUserObject();
        return null;
    }

    /** Set the selected node, identified by the user object
     *  @param obj object, which should be selected
     */
    public void setSelectedObject(Object obj) {
        if (obj == null)
            return;
        DefaultMutableTreeNode node = (DefaultMutableTreeNode) findNode(topnode, obj);
        if (node == null)
            return;
        if (node == (DefaultMutableTreeNode) comptree.getLastSelectedPathComponent())
            return;
        comptree.setSelectionPath(new TreePath(node.getPath()));
    }
    
    
    /** get the parent object of a given object 
     *  @param obj user object to look for
     */
    public Object getParentObject(Object obj) {
        if (obj == null)
            return null;
        DefaultMutableTreeNode node = (DefaultMutableTreeNode) findNode(topnode, obj);
        if (node == null)
            return null;
        node = (DefaultMutableTreeNode) node.getParent();
        if (node == null)
            return null;
        return ((JVPanelTreeComponent)node.getUserObject()).getObject();
    }
    
    /** find a node identified by the node user object
     *  @param userobj user object to look for
     */
    private TreeNode findNode (TreeNode node, Object userobj) {
        if (((JVPanelTreeComponent)((DefaultMutableTreeNode)node).getUserObject()).getObject() == userobj)
            return node;
        for (int i = 0; i < node.getChildCount(); i++) {
            TreeNode subnode = findNode(node.getChildAt(i), userobj);
            if (subnode != null)
                return subnode;
        }
        return null;
    }
    
    /** Add tree selection listener */
    public void addTreeSelectionListener(TreeSelectionListener listener) {
        comptree.addTreeSelectionListener(listener);
    }

    /** Remove tree selection listener */
    public void removeTreeSelectionListener(TreeSelectionListener listener) {
        comptree.removeTreeSelectionListener(listener);
    }
}
